! -*- Mode: f90 -*-
PROGRAM mosaic
   ! -------------------------------------------------------------------------------
   !
   !> @brief Generates interpolation weights and adresses for Oasis Coupler, Mozaic interpolation
   !
   !>
   !> Oceanic model : OPA (C Grid)
   !> Atmospheric Model : LMD5, LMDz or spectral model with Gaussian grid
   !>
   !> SVN          :
   !> $Author$
   !> $Date$
   !> $Revision$
   !> $Id$
   !>
   !> $Log$
   !>
   !>
   !> -------------------------------------------------------------------------------
   !>
   !> http://www.ipsl.jussieu.fr/~omamce/IPSLCM4/MOZAIC
   !>
   !> -------------------------------------------------------------------------------
   !>
   !> Warning, to install, configure, run, use any of Olivier Marti's
   !> software or to read the associated documentation you'll need at least
   !> one (1) brain in a reasonably working order. Lack of this implement
   !> will void any warranties (either express or implied). 
   !> O. Marti assumes no responsability for errors, omissions,
   !> data loss, or any other consequences caused directly or indirectly by
   !> the usage of his software by incorrectly or partially configured
   !> personal.
   !>
   !> 1) Read mask, grid and areas files, at OASIS format
   !>     Convention for mask are those of Oasis : 0 for sea, 1 for land
   !>
   !> 2) Generates interpolation weights
   !>
   !> 3) Write them at Oasis format
   !>
   !> 4) Write some diagnostic fields in NetCDF format
   !>
   !>
   !> @author Olivier MARTI, june 1996
   !
   !> Modification : Olivier MARTI, may 1997 : LMDZ case, from L. Li
   !>                Eric GUILYARDI, may 1999 : Gaussian Grid + ORCA
   !>                Olivier MARTI and Jacques BELLIER, april 2000 : use of polygon matching to compute weights
   !>
   ! --------------------------------------------------------------------------------
   ! LSCE, CEA-Saclay
   ! --------------------------------------------------------------------------------
   !
   !
   USE modeles
   USE mod_inipar
   USE mod_locio
   USE mod_neuf
   USE bords
   USE mod_lbc
   USE mod_wripoly
   USE mod_polyw_a
   USE mod_polyw_o
   USE formula
   USE mod_lecdata
   USE mod_prih
   USE mod_norma
   USE mod_rectifie
   USE math
   USE mod_inter
   USE mod_wri_wei
   USE fliocom
   USE getincom
   USE errioipsl
   USE mod_polom
   USE mod_proj
   USE fliocom
   !
   IMPLICIT NONE
   !
   REAL (kind=rl) :: zzz
   REAL (kind=rd) :: zs, zs1, zs2
   INTEGER (kind=il) :: ja, jo, jai, jaj, joi, joj, jno, jna, kn ! Dummy indexes
   INTEGER (kind=il), DIMENSION (2) :: jind
   CHARACTER (len = 80) :: cfname4, cfname8
   !
   INTEGER, DIMENSION (:,:), ALLOCATABLE :: mt1, mt2, mt3
   !
   ! NetCDF
   CHARACTER (len = 50) :: cldiag_o2a ! NetCDF file name for fields on atmospheric grid (diagnostics)
   CHARACTER (len = 50) :: cldiag_a2o ! NetCDF file name for fields on oceanic grid  (diagnostics)
   CHARACTER (len = 50) :: clw_o2a, clw_a2o, clw_a2o_mct, clw_o2a_mct
   CHARACTER (len = 25) :: mo_name, o2a_name, a2o_name
   CHARACTER (len =  1) :: c_i, c_r
   INTEGER :: il_mask
   CHARACTER (len=10) :: cmodel
   !
   REAL (kind=rd), DIMENSION (:), ALLOCATABLE :: xa, ya, xo, yo
   INTEGER (kind=il) :: ierr
   INTEGER (kind=il), DIMENSION (:,:), ALLOCATABLE :: itmp
   CHARACTER (len=80) :: clarg
   INTEGER :: narg
   !>
   !! Lecture de la ligne de comande
   narg = iargc()
   IF ( narg > 0 ) THEN
      CALL getarg ( 1, clarg)
      SELECT CASE (TRIM(clarg))
      CASE Default
         IF ( INDEX ( TRIM(clarg), '.def' ) /= 0 ) THEN
            WRITE (nout,*) 'Lecture des parametres dans ', TRIM(clarg)
            CALL getin_name ( TRIM (clarg) )
         ELSE
            WRITE (nout,*) 'Lecture des parametres dans run.def'
         ENDIF
      END SELECT
   END IF
   !
   CALL inipar
   CALL alloc_modeles
   !
   !> @par Initialisation
   !
   ngrd = 11 ; nsrf = 12 ; nmsk = 13; nchk = 16 ; ndeb = 9 ; nbug1 = 10
   nwei4o2a = 21 ; nwei8o2a = 22 ; nwei4a2o = 23 ; nwei8a2o = 24
   ! Diagnostics files
   !
   nout = 6
   IF (nout /= 6 .AND. nout /= 0) THEN 
      OPEN (unit = nout, file = 'mosaic' // TRIM(c_suffix) // '.out', position = 'REWIND', action = 'WRITE', status = 'replace')
   END IF
   !>

   WRITE (nout,*) 'm2ai ', MINVAL (m2ai), MAXVAL (m2ai)
   WRITE (nout,*) 'm2aj ', MINVAL (m2aj), MAXVAL (m2aj) 
   !
   ndebug = 2
   !
   IF (lwra2o) THEN
      WRITE (unit = nout, fmt = *) ' Compute weights and adresses for interpolation atm -> oce '
   ELSE
      WRITE (unit = nout, fmt = *) ' Nothing for atm -> oce '
   ENDIF
   IF (lwro2a) THEN
      WRITE (unit = nout, fmt = *) ' Compute weights and adresses for interpolation oce -> atm '
   ELSE
      WRITE (unit = nout, fmt = *) ' Nothing for oce -> atm '
   ENDIF
   !
   WRITE (nout,*) 'epsfrac : ', epsfrac
   !
   ka2o = 0 ; wa2o = 0.0_rl ; ko2a = 0 ; wo2a = 0.0_rl
   !
   !> Read coordinates of all models
   !
   IF (l_dryrun .AND. lev_dry >= 1 ) THEN 
      CALL lecdata (lecweights = .TRUE. , leco2amask =.FALSE.)
   ELSE
      CALL lecdata (lecweights = .FALSE., leco2amask =.FALSE.)
   ENDIF
   !
   !> Set periodicity
   CALL lbc (xolont, noperio, 'T', jpoi, jpoj)
   CALL lbc (xolonu, noperio, 'U', jpoi, jpoj)
   CALL lbc (xolonv, noperio, 'V', jpoi, jpoj)
   CALL lbc (xolonf, noperio, 'F', jpoi, jpoj)
   CALL lbc (xolatt, noperio, 'T', jpoi, jpoj)
   CALL lbc (xolatu, noperio, 'U', jpoi, jpoj)
   CALL lbc (xolatv, noperio, 'V', jpoi, jpoj)
   CALL lbc (xolatf, noperio, 'F', jpoi, jpoj)
   CALL lbc (xosrft, noperio, 'T', jpoi, jpoj)
   CALL lbc (iomskt, noperio, 'T', jpoi, jpoj)
   !
!-$$   CALL lbc (xalatt, naperio, 'T', jpai, jpaj)
!-$$   CALL lbc (xalatu, naperio, 'U', jpai, jpaj)
!-$$   CALL lbc (xalatv, naperio, 'V', jpai, jpaj)
!-$$   CALL lbc (xasrft, naperio, 'T', jpai, jpaj)
!-$$   CALL lbc (iamskt, naperio, 'T', jpai, jpaj)
   !>
   CALL rectifie
   !>
   CALL bord_a
   !
   CALL bord_o
   !
   CALL bord_oa
   !
   !> Surfaces par les polygones
   ALLOCATE (xa (jpae), STAT=ierr) ; CALL chk_allo (ierr, 'xa in mosaic')
   ALLOCATE (ya (jpae), STAT=ierr) ; CALL chk_allo (ierr, 'ya in mosaic')
   !
   IF (.NOT. l_dryrun .AND. l_recalc_a) THEN 
      WRITE (nout,*) 'Recalcul les surfaces atm'
      DO ja = 1, jpan
         DO kn = 1, jpae
            CALL proj (xa_ed (ja,kn), ya_ed (ja,kn), REAL(xa_ed(ja,jpae),rd), REAL(ya_ed(ja,jpae),rd), xa(kn), ya(kn), 2, ierr)
         ENDDO
         CALL polom (xa, ya, xa, ya, zs1, zs2, zs)
         IF (MAX (zs1, zs2) .GT. MIN (zs1, zs2) + zeps .OR. ABS(zs-1.0_rl) > zeps) THEN 
            WRITE (nout,*) 'pb surface atm : ', ja, mai(ja), maj(ja), zs1, zs2, zs
         ELSE
            xasrft_pol (ja) = zs1
         END IF
      END DO
   END IF
   !
   WRITE (nout,*) 'Perio sur atm'
   CALL lbc (xasrft_pol, naperio, 'T', jpai, jpaj)
   !
   WRITE (nout, *) 'Points T Atmosphere'
   WRITE (nout, *) ' Longitude Atm '
   WRITE (nout, '(10F7.1)') (xalont(m1a(jai,jpaj/2)),jai=1,jpai)    ! xalont (jpan/2+1:jpan/2+jpai)
   WRITE (nout, *) ' Latitude Atm '
   WRITE (nout, '(10F7.1)') (xalatt(m1a(jpai/2,jaj)),jaj=1,jpaj)    ! xalatt (1:jpan:jpai)
   WRITE (nout, *) 'surface Atm'
   CALL prihre(RESHAPE(xasrft,(/jpai,jpaj/)), 0.0_rl, nout)!-$$   WRITE (nout, *) 'Points east Atm'

   WRITE (nout, *) 'Mask Atm'
   DO jaj = jpaj, 1, -1
      WRITE (nout, '(144I1)' ) (iamskt (m1a (jai, jaj)), jai = 1, jpai)
   END DO
   WRITE (nout, *) 'Mask Atm perio'
   DO jaj = jpaj, 1, -1
      WRITE (nout, '(144I1)' ) (iamskp (m1a (jai, jaj)), jai = 1, jpai)
   END DO
   WRITE (nout, *) 'Mask Oce'
   DO joj = jpoj, 1, -1
      WRITE (nout, '(1I5, "  ", 182I1)') joj, (iomskt (m1o (joi, joj)), joi = 1, jpoi    )
   END DO
   WRITE (nout, '(1I5, 18I10)')    joj, ( joi                   , joi = 1, jpoi, 10)
   WRITE (nout, *) 'Mask Oce Perio'
   DO joj = jpoj, 1, -1
      WRITE (nout, '(1I5, "  ", 182I1)') joj, (iomskp (m1o (joi, joj)), joi = 1, jpoi    )
   END DO
   WRITE (nout, '(1I5, 18I10)')    joj, ( joi                   , joi = 1, jpoi, 10)

   !
   ALLOCATE (xo (jpoe), STAT=ierr) ; CALL chk_allo (ierr, 'xo in mosaic')
   ALLOCATE (yo (jpoe), STAT=ierr) ; CALL chk_allo (ierr, 'yo in mosaic')
   !
   IF (.NOT. l_dryrun .AND. l_recalc_o) THEN
      WRITE (nout,*) 'Recalcul des surfaces oce'
      ReOce: DO jo = 1, jpon
         IF ( iomskp(jo) == 0 .AND. iomskt(jo) == 0 ) THEN 
            ! WRITE (nout,*) jo, xo_ed (jo,jpoe), yo_ed (jo,jpoe)
            DO kn = 1, jpoe
               CALL proj (xo_ed (jo,kn), yo_ed (jo,kn), REAL(xo_ed(jo,jpoe),rd), REAL(yo_ed(jo,jpoe),rd), xo(kn), yo(kn), 2, ierr)
            ENDDO
            CALL polom (xo, yo, xo, yo, zs1, zs2, zs)
            IF (MAX (zs1, zs2) .GT. MIN (zs1, zs2) + zeps ) THEN 
               WRITE (nout,*) 'pb 1 surface oce : ', jo, iomskp(jo), iomskt(jo), m2oi(jo), m2oj(jo), &
                  & xolont(jo), xolatt(jo), zs1, zs2, ABS(zs1-zs2), ABS(zs1-zs2)/MAX(zs1,zs2),zs, zs-1.0_rl
            END IF
            IF (ABS(zs-1.0_rl) > zeps) THEN 
               WRITE (nout,*) 'pb 2 surface oce : ', jo, iomskp(jo), iomskt(jo), m2oi(jo), m2oj(jo), &
                  & xolont(jo), xolatt(jo), zs1, zs2, ABS(zs1-zs2), ABS(zs1-zs2)/MAX(zs1,zs2),zs, zs-1.0_rl
            END IF
            xosrft_pol (jo) = zs1
         END IF
      END DO ReOce
   END IF
!-$$   xosrft_pol = xosrft
   
   CALL lbc (xosrft_pol, noperio, 'T', jpoi, jpoj)
   !
   WRITE (nout, *) 'Points T Ocean'
   WRITE (nout, *) ' Longitudes Oce '
   WRITE (nout, '(10F7.1)') xo_ed (1: jpoi, jpoe)
   DO jno = 1, jpoe-1
      WRITE (nout, *) 'Longitudes points Oce ', jno
      WRITE (nout, '(10F7.1)') xo_ed (1: jpoi, jno)
      WRITE (nout, *) 'Latitudes points Oce ', jno
      WRITE (nout, '(10F7.1)') yo_ed (1: jpon: jpoj, jno)
   ENDDO
   !
   DO jna = 1, jpae-1
      WRITE (nout, *) 'Longitudes points Atm ', jna
      WRITE (nout, '(10F7.1)') xa_ed (1: jpai, jna)
      WRITE (nout, *) 'Latitudes points Atm ', jna
      WRITE (nout, '(10F7.1)') ya_ed (1: jpan: jpaj, jna)
   ENDDO
   !
   WRITE (nout,*) 'total surface oce ', SUM(xosrft*REAL(1-iomskp,rl))
   WRITE (nout,*) 'total surface atm ', SUM(xasrft*REAL(1-iamskp,rl))
   !
   IF (l_dryrun) THEN
      SELECT CASE (lev_dry)
      CASE (0)
         WRITE (nout,*) "Remplissage bidon des champs pour tester les I/O"
!-$$         jma2o = jpa2o ; jmo2a = jpo2a
         nvo = jma2o   ; nva   = jmo2a
         CALL RANDOM_SEED
         CALL RANDOM_NUMBER (wa2o)
         CALL RANDOM_NUMBER (wo2a)
         
         WRITE (nout,*) 'wa2o : ', MINVAL (wa2o), MAXVAL (wa2o)
         WRITE (nout,*) 'wo2a : ', MINVAL (wo2a), MAXVAL (wo2a)
         
         ka2o (:,:) = NINT ( wa2o(:,:) * REAL(jpan,rl) )
         ko2a (:,:) = NINT ( wo2a(:,:) * REAL(jpon,rl) )

         WRITE (nout,*) 'ka2o : ', MINVAL (ka2o), MAXVAL (ka2o)
         WRITE (nout,*) 'ko2a : ', MINVAL (ko2a), MAXVAL (ko2a)

         wa2o(:,:) = ( 0.5 + wa2o(:,:) ) / REAL ( jma2o, KIND=rl)
         wo2a(:,:) = ( 0.5 + wo2a(:,:) ) / REAL ( jmo2a, KIND=rl)

         WRITE (nout,*) 'wa2o : ', MINVAL (wa2o), MAXVAL (wa2o)
         WRITE (nout,*) 'wo2a : ', MINVAL (wo2a), MAXVAL (wo2a)

         !! Masquage
         DO ja = 1, jpan
            IF ( iamskt (ja) == 1 ) THEN 
               wo2a (:, ja) = 0.0_rl ; ko2a (:, ja) = 0_il
            END IF
         END DO
         DO jo = 1, jpon
            IF ( iomskt (jo) == 1 ) THEN
               wa2o (:, jo) = 0.0_rl ; ka2o (:, jo) = 0_il
            END IF
         END DO

         DO ja = 1, jpan
            DO jna = 1, jpo2a
               jo = ko2a (jna, ja)
               IF (jo > 0) THEN 
                  IF (iomskt (jo) == 1) THEN
                     wo2a (jna, ja) = 0.0_rl ; ko2a (jna, ja) = 0_il
                  END IF
               END IF
            END DO
         END DO

         DO jo = 1, jpon
            DO jno = 1, jpa2o
               ja = ka2o (jno, jo)
               IF (ja > 0) THEN 
                  IF (iamskt (ja) == 1) THEN
                     wa2o (jno, jo) = 0.0_rl ; ka2o (jno, jo) = 0_il
                  END IF
               END IF
            END DO
         END DO
         !! Normalisation
         DO jo = 1, jpon
            wa2o(:,jo) = wa2o (:,jo) / SUM ( wa2o (:,jo))
         END DO
         DO ja = 1, jpan
            wo2a(:,ja) = wo2a (:,ja) / SUM ( wo2a (:,ja))
         END DO

         CALL gather ('o2a')
         !CALL inter_o2a (REAL (1-iomskt, KIND = rl), o2amask)
         !CALL rem_small ('o2a')
         CALL gather ('a2o')
         
         WRITE (nout,*) 'ka2o : ', MINVAL (ka2o), MAXVAL (ka2o)
         WRITE (nout,*) 'ko2a : ', MINVAL (ko2a), MAXVAL (ko2a)
         
         WRITE (nout,*) 'wa2o : ', MINVAL (wa2o), MAXVAL (wa2o)
         WRITE (nout,*) 'wo2a : ', MINVAL (wo2a), MAXVAL (wo2a)
         
      CASE (1)
         WRITE (nout,*) "Utilisation des poids lus"
      END SELECT
   ELSE
      WRITE (nout,*) 'Appel de polyw_a'
      CALL polyw_a (lo2a = .TRUE., la2o = .TRUE.)
      WRITE (nout,*) 'Appel de gather(o2a) apres polyw_a'
      CALL gather ('o2a')
      WRITE (nout,*) 'Appel de gather(a2o) apres polyw_a'
      CALL gather ('a2o')
   ENDIF
   !
      
   !> Checking weights
   WRITE (unit = nout, fmt = '("Neighbors o2a     : ", 3I12 )') MINVAL (nva, nva /=0), MAXVAL (nva), jmo2a
   WRITE (unit = nout, fmt = '("Index o2a         : ", 2I12 )') MINVAL (ko2a, ko2a /= 0), MAXVAL (ko2a)
   jind = MINLOC(wo2a, wo2a /= 0.0_rl) ; ja = jind(2) ; jna = jind(1); jai = m2ai(ja) ; jaj = m2aj(ja)
   WRITE (unit = nout, fmt = '("Weights o2a mini : ", 2I8, 2I4, 1E13.4, 3I8)') &
      & ja, jna, jai, jaj, wo2a(jna,ja), ko2a(jna,ja), m2oi(ko2a(jna,ja)), m2oj(ko2a(jna,ja))
   jind = MAXLOC(wo2a, wo2a /= 0.0_rl) ; ja = jind(2) ; jna = jind(1); jai = m2ai(ja) ; jaj = m2aj(ja)
   WRITE (unit = nout, fmt = '("Weights o2a maxi : ", 2I8, 2I4, 1E13.4, 3I8)') &
      & ja, jna, jai, jaj, wo2a(jna,ja), ko2a(jna,ja), m2oi(ko2a(jna,ja)), m2oj(ko2a(jna,ja))
   !> Ocean mask interpolated to atm
   CALL inter_o2a (REAL (1-iomskt, KIND = rl), o2amask)
!-$$   WRITE (nout,*) 'Ocean mask interpolated to atm before normalisation'
!-$$   CALL prihre(RESHAPE(o2amask,(/jpai,jpaj/)),10.0_rl,INT(nout,kind=i_4))
   !
   !> Classes
   WRITE (nout, *) 'Low sea fraction before normalization'
   DO kn = 1, 16
      zzz = 10.0_rl**(-kn)
      WRITE (nout, *) 'Classe : ', kn, COUNT (o2amask .LE. zzz .AND. o2amask .GT. zzz/10.0_rl)
   END DO
   WRITE (nout, *) 'High sea fraction'
   DO kn = 1, 16
      zzz = 10.0_rl**(-kn)
      WRITE (nout, *) 'Classe : ', kn, COUNT (1.0_rl-o2amask .LE. zzz .AND. 1.0_rl-o2amask .GT. zzz/10.0_rl)
   END DO
   !> Remove small fractions
   CALL rem_small ('o2a')
   !> Recalcule le masque. A faire avant la normalisation
   CALL inter_o2a (REAL (1-iomskt, KIND = rl), o2amask)
   ! Petit bug ...
   !IF ( catyp == 'lmdz' ) o2amask (jpan-jpai+1:jpan) = 1.0_rl
   !
!-$$   WRITE (nout, *) 'ATM : Sea fraction < 0'
!-$$   DO ja = 1, jpan
!-$$      IF (o2amask (ja) < 0.0_rl) THEN 
!-$$         WRITE (nout,*) ja, m2ai (ja), m2aj (ja), o2amask (ja)
!-$$      ENDIF
!-$$   ENDDO
!-$$   WRITE (nout, *) 'ATM : Sea fraction > 1'
!-$$   DO ja = 1, jpan
!-$$      IF (o2amask (ja) > 1.0_rl+eps) THEN 
!-$$         WRITE (nout,*) ja, m2ai (ja), m2aj (ja), o2amask (ja)
!-$$      ENDIF
!-$$   ENDDO

   o2amask_i_int (:) = 1
   o2amask_i_ext (:) = 1
   WHERE ( o2amask (:) >  0.0_rl         ) o2amask_i_int = 0
   WHERE ( o2amask (:) > 1.0_rl-epsfrac ) o2amask_i_ext = 0
 
   !
   WRITE (nout,*) 'Appel de gather a2o avant verif'
   CALL gather ('a2o')
   !
   !> Checking weights
   WRITE (unit = nout, fmt = '("Neighbors a2o     : ", 3I9 )') MINVAL (nvo, nvo /=0), MAXVAL (nvo), jma2o
   WRITE (unit = nout, fmt = '("Index a2o         : ", 2I9 )') MINVAL (ka2o, ka2o /= 0), MAXVAL (ka2o)
   jind = MINLOC(wa2o, wa2o /= 0.0_rl) ; jo = jind(2) ; jno = jind(1); joi = m2oi(jo) ; joj = m2oj(jo)
   WRITE (unit = nout, fmt = '("Weights a2o mini : ", 2I8, 2I4, 1E13.4, 3I6)') &
      & jo, jno, joi, joj, wa2o(jno,jo), ka2o(jno,jo), m2ai(ka2o(jno,jo)), m2aj(ka2o(jno,jo))
   jind = MAXLOC(wa2o, wa2o /= 0.0_rl) ; jo = jind(2) ; jno = jind(1); joi = m2oi(jo) ; joj = m2oj(jo)
   WRITE (unit = nout, fmt = '("Weights a2o maxi : ", 2I8, 2I4, 1E13.4, 3I6)') &
      & jo, jno, joi, joj, wa2o(jno,jo), ka2o(jno,jo), m2ai(ka2o(jno,jo)), m2aj(ka2o(jno,jo))
   !> Atm mask interpolated to oce
   CALL inter_a2o (REAL (1-iamskt, KIND = rl), a2omask)
   WRITE (nout,*) 'Atm mask interpolated to oce before normalization'
   !$$$   CALL prihre(RESHAPE(a2omask,(/jpoi,jpoj/)),10.0_rl,INT(nout,kind=i_4))
   !> Remove small fractions
   CALL rem_small ('a2o')
   CALL inter_a2o (REAL (1-iamskt, KIND = rl), a2omask)
   !
!-$$   WRITE (nout, *) 'OCE : Sea fraction < 0'
!-$$   DO jo = 1, jpon
!-$$      IF (a2omask(jo) < 0.0) THEN 
!-$$         WRITE (nout,*) jo, m2oi(jo), m2oj(jo), o2amask(jo)
!-$$      ENDIF
!-$$   ENDDO
!-$$   WRITE (nout, *) 'OCE : Sea fraction > 1'
!-$$   DO jo = 1, jpon
!-$$      IF (a2omask(jo) > 1.0+eps) THEN 
!-$$         WRITE (nout,*) jo, m2oi(jo), m2oj(jo), a2omask(jo)
!-$$      ENDIF
!-$$   ENDDO
   ! Normalise
   IF (TRIM (catyp) == 'lmdz' ) THEN 
      WRITE (unit=nout, FMT=*) 'Calling normalization '
      CALL norma (knorma2o = norma2o, knormo2a = normo2a)
   END IF
   !
   WRITE (unit=nout,fmt='("Maxval de sum(wo2a)-1 ", 1E15.4)') MAXVAL (SUM (wo2a(:,:),DIM=1)) -1.0_rl
   !
   WRITE (unit = nout, fmt = '("Surface atmos               : ", 1PE15.4)') &
      DOT_PRODUCT (xasrft, REAL(1-iamskp,kind=rl))
   WRITE (unit = nout, fmt = '("Surface ocean               : ", 1PE15.4)') &
      DOT_PRODUCT (xosrft, REAL(1-iomskp,kind=rl))
   !
   WRITE (unit = nout, fmt = '("Oce->Atm : ", 1I9, " non-null adresses, ", 1I12, " non-null weights ")') & 
      & COUNT (wo2a > 0.0_rl), COUNT (ko2a > 0)
   WRITE (unit = nout, fmt = '("Atm->Oce : ", 1I9, " non-null adresses, ", 1I12, " non-null weights ")') &
      & COUNT (wa2o > 0.0_rl), COUNT (ka2o > 0)
   !
   xsurfa = DOT_PRODUCT (xasrft*REAL(1-iamskp,kind=rl), 1.0_rl - o2amask)
   xsurfo = DOT_PRODUCT (xosrft, REAL((1-iomskt)*(1-iomskp), kind=rl))
   WRITE (nout, fmt= '("Ocean surface on ocean and atmosphere : ", 2(1PE20.6))') xsurfo, xsurfa
   !
   !> Write Weights for OASIS
   !
   mo_name = 'mozaic'
   o2a_name = "o2a" ; a2o_name = "a2o"
   WRITE (unit = c_r, fmt = '(1i1)') rk_out
   IF (lwro2a) THEN 
      IF (l_wei_i4) THEN 
         WRITE (unit = c_i, fmt = '(1i1)') i_4
         cfname4 = TRIM(mo_name) // ".wo2a"  // TRIM(c_suffix) // ".i" // TRIM(c_i) // ".r" // TRIM(c_r) 
         WRITE (nout, *) 'Poids i4 o -> a: ', TRIM(cfname4)
         OPEN (unit = nwei4o2a, file = TRIM(cfname4), form = 'unformatted', status = 'replace', action = 'write')
      ENDIF
      !
      IF (l_wei_i8) THEN
         WRITE (unit = c_i, fmt = '(1i1)') i_8
         cfname8 = TRIM(mo_name) // ".wo2a" // TRIM(c_suffix) // ".i" // TRIM(c_i)  // ".r" // TRIM(c_r) 
         WRITE (nout, *) 'Poids i8 : ', TRIM(cfname8)
         OPEN (unit = nwei8o2a, file = TRIM(cfname8), form = 'unformatted', status = 'replace', action = 'write')
      ENDIF
      !
      cldiag_o2a  = TRIM(o2a_name) // '.diag'
      clw_o2a     = TRIM(mo_name)  // '.wo2a'
      clw_o2a_mct = "rmp_"//TRIM(comod_t)//"_to_"//TRIM(camod_t)//"_MOSAIC"
      CALL wri_weights_o2a (cldiag_o2a, clw_o2a, clw_o2a_mct, "1", l_fulldiag=.FALSE., &
         & lo_src_grid_frac=.FALSE., lo_dst_grid_frac=.FALSE.)
      CLOSE (unit = nwei8o2a) ; CLOSE (unit = nwei4o2a)

   ENDIF

   IF (lwra2o) THEN 
      !$$$      WRITE (nout,*) 'Ocean mask interpolated to atm, before writing'
      !$$$      CALL prihre(RESHAPE(o2amask,(/jpai,jpaj/)),10.0_rl,INT(nout,kind=i_4))
      IF (l_wei_i4) THEN
         WRITE (unit = c_i, fmt = '(1i1)') i_4
         cfname4 = TRIM(mo_name) // ".wa2o"  // TRIM(c_suffix) // ".i" // TRIM(c_i) // ".r" // TRIM(c_r)
         WRITE (nout, *) 'Poids i4 a -> o: ', TRIM(cfname4)
         OPEN (unit = nwei4a2o, file = TRIM(cfname4), form = 'unformatted', status = 'replace', action = 'write')
      END IF

      IF (l_wei_i8) THEN
         WRITE (unit = c_i, fmt = '(1i1)') i_8
         cfname8 = TRIM(mo_name) // ".wa2o" // TRIM(c_suffix) // ".i" // TRIM(c_i) // ".r" // TRIM(c_r) 
         WRITE (nout, *) 'Poids i8 : ', TRIM(cfname8)
         OPEN (unit = nwei8a2o, file = TRIM(cfname8), form = 'unformatted', status = 'replace', action = 'write')
      ENDIF

      cldiag_a2o  = TRIM(a2o_name) // '.diag'
      clw_a2o     = TRIM(mo_name)  // '.wa2o'
      clw_a2o_mct = "rmp_"//TRIM(camod_t)//"_to_"//TRIM(comod_t)//"_MOSAIC"
      CALL wri_weights_a2o (cldiag_a2o, clw_a2o, clw_a2o_mct, "2", l_fulldiag=.FALSE., &
         & lo_src_grid_frac=.FALSE., lo_dst_grid_frac=.FALSE.)

      IF (l_wei_i4) CLOSE (unit = nwei8a2o) 
      IF (l_wei_i8) CLOSE (unit = nwei4a2o)

   ENDIF

   !$$$   IF (lwro2a) THEN 
   !$$$      !> Compute o2a weights for interpolating fluxes
   !$$$      !
   !$$$      !
   !$$$      !> Normalize weights
   !$$$      WRITE (nout, *) 'Calling normalization '
   !$$$      CALL norma (knorma2o = 0_il, knormo2a = 2_il)
   !$$$      !
   !$$$      WRITE (unit = cfname4, fmt = '("mozaic.wo2a.flx.i", 1I1, ".r", 1I1)') i_4, rk_out
   !$$$      OPEN (unit = nwei4o2a, file = TRIM(cfname4), form = 'unformatted', status = 'replace', action = 'write')
   !$$$      WRITE (nout, *) 'Poids i4 o -> a: ', TRIM(cfname4)
   !$$$
   !$$$      WRITE (unit = cfname8, fmt = '("mozaic.wo2a.flx.i", 1I1, ".r", 1I1)') i_8, rk_out
   !$$$      OPEN (unit = nwei8o2a, file = TRIM(cfname8), form = 'unformatted', status = 'replace', action = 'write')
   !$$$      WRITE (nout, *) 'Poids i8 : ', TRIM(cfname8)
   !$$$
   !$$$      cldiag_o2a = 'o2a_flx_diag' ; clw_o2a    = 'mozaic.wo2a.flx'
   !$$$
   !$$$      nwei4o2a = 56 ; nwei8o2a = 57
   !$$$      CALL wri_weights_o2a (cldiag_o2a, clw_o2a, "6")
   !$$$
   !$$$      CLOSE (unit = nwei8o2a)
   !$$$      CLOSE (unit = nwei4o2a)
   !$$$
   !$$$      !
   !$$$   ENDIF
   !
   !
   CALL flioclo
   WRITE (unit = nout, fmt = *) 'That''s all folks ! '
   !
   STOP
CONTAINS
   !
   SUBROUTINE perio_w (clname)
      CHARACTER (len=*), INTENT (in) :: clname
      !
      SELECT CASE (TRIM(clname))
      CASE ('a2o')
         !> Periodicity ocean
         jma2o = MAXVAL (nvo)
         DO jno = 1, jma2o
            CALL lbc (wa2o (jno,:) , noperio, 'T', jpoi, jpoj)
            CALL lbc (ka2o (jno,:) , noperio, 'T', jpoi, jpoj)
         END DO
      CASE ('o2a')
         !> Periodicity atm
         jmo2a = MAXVAL (nva)
         DO jna = 1, jmo2a
            CALL lbc (wo2a (jna,:) , naperio, 'T', jpai, jpaj)
            CALL lbc (ko2a (jna,:) , naperio, 'T', jpai, jpaj)
         END DO
         !
      END SELECT
   END SUBROUTINE perio_w
   !
   SUBROUTINE gather (clname)
      !
      CHARACTER (len=*), INTENT (in) :: clname
      INTEGER (kind=il), DIMENSION (:), ALLOCATABLE :: ktemp
      REAL    (kind=rl), DIMENSION (:), ALLOCATABLE :: wtemp
      INTEGER (kind=il) :: ierr, jn, jon1, jon2, ja1, ja2
      !
     
      !
      SELECT CASE (TRIM(clname))
      CASE ('o2a')
         !
         CALL perio_w (clname)
         !
         ALLOCATE (ktemp (jpo2a), STAT=ierr) ; CALL chk_allo (ierr, 'ktemp (jpo2a)', crout='mosaic, gather o2a')
         ALLOCATE (wtemp (jpo2a), STAT=ierr) ; CALL chk_allo (ierr, 'wtemp (jpo2a)')
         DO ja = 1, jpan
            ktemp (:) = ko2a (:,ja) ; wtemp (:) = wo2a (:,ja)
            ko2a (:,ja) = 0_il  ; wo2a (:,ja) = 0.0_rl
            kn = 1
            DO jna = 1, jmo2a
               IF ( ktemp (jna) /= 0_il ) THEN
                  ko2a (kn, ja) =  ktemp (jna)
                  wo2a (kn, ja) =  wtemp (jna)
                  kn = kn + 1
               END IF
            END DO
         END DO
         nva = COUNT (ko2a /= 0, DIM = 1)
         IF ( MAXVAL (nva) > jmo2a ) THEN
            WRITE (nout, FMT=*) 'Attention: augmentation de jmo2a dans gather o2a'
            STOP
         ELSE
            jmo2a = MAXVAL (nva)
         END IF
         WRITE (unit = nout, fmt = *) 'Nombre de voisins o->a : ', jmo2a
         DEALLOCATE (ktemp) ; DEALLOCATE(wtemp)
         !
      CASE ('a2o')
         IF ( naperio == -1 ) THEN 
            ! Pointage vers le point pole
            DO jo = 1, jpon
               DO jn = 1, jma2o
                  IF ( m2aj (ka2o (jn, jo)) == 1    ) ka2o (jn, jo) = m1a ( 1,  1  )
                  IF ( m2aj (ka2o (jn, jo)) == jpaj ) ka2o (jn, jo) = m1a ( 1, jpaj)
               END DO
            END DO
         END IF
         !
         ! Si deux adresses identiques, simplification et somme des poids
         DO jo = 1, jpon
            DO jon1 = 1, jma2o
               ja1 = ka2o (jon1, jo)
               IF ( ja1 == 0 ) CYCLE
               DO  jon2 = 1, jpa2o
                  IF ( jon1 == jon2 ) CYCLE
                  ja2 = ka2o (jon2, jo)
                  IF ( ja1 == ja2 ) THEN
                     wa2o (jon1, jo) = wa2o (jon1, jo) + wa2o (jon2, jo)
                     wa2o (jon2, jo) = 0.0_rl
                     ka2o (jon2, jo) = 0
                  END IF
               END DO
            END DO
         END DO
         !
         CALL perio_w (clname)      
         !
         ALLOCATE (ktemp (jpa2o), STAT=ierr) ; CALL chk_allo (ierr, 'ktemp (jpa2o)', crout='mosaic, gather a2o')
         ALLOCATE (wtemp (jpa2o), STAT=ierr) ; CALL chk_allo (ierr, 'wtemp (jpa2o)')
         DO jo = 1, jpon
            ktemp = ka2o(:,jo) ; wtemp = wa2o(:,jo)
            ka2o (:,jo) = 0_il  ; wa2o (:,jo) = 0.0_rl
            kn = 1_il
            DO jno = 1, jma2o
               IF ( ktemp (jno) /= 0_il ) THEN
                  ka2o (kn, jo) =  ktemp (jno)
                  wa2o (kn, jo) =  wtemp (jno)
                  kn = kn + 1_il
               END IF
            END DO
         END DO
         nvo = COUNT (ka2o /= 0, DIM = 1)
         IF ( MAXVAL (nvo) > jma2o) THEN
            WRITE (nout,*) 'Attention: augmentation de nvo dans gather'
            STOP
         ELSE
            jma2o = MAXVAL (nvo)
         ENDIF
         WRITE (unit = nout, fmt = *) 'Nombre de voisins a->o : ', jma2o
         DEALLOCATE (ktemp) ; DEALLOCATE (wtemp)
         !
      END SELECT

      !
   END SUBROUTINE gather
   !
!-$$   SUBROUTINE pole_a2o
!-$$      !
!-$$      IMPLICIT NONE
!-$$      !
!-$$      DO jn = 1, jma2o
!-$$         DO jo = 
!-$$         
   !
   SUBROUTINE rem_small (clname)
      !
      CHARACTER (len=*), INTENT (in) :: clname
      REAL (kind=rl) :: zzs
      INTEGER :: nn_count
      !
      nn_count = 0
      SELECT CASE (TRIM(clname))
      CASE ( 'a2o' )
         !> a -> o
         DO jo = 1, jpon
            zzs = SUM (wa2o (:,jo))
            DO jno = 1, jma2o
               ja = ka2o (jno, jo)
               IF (ja > 0) THEN 
                  IF (wa2o (jno, jo) < epsfrac) THEN 
                     wa2o (jno, jo) = 0.0_rl
                     ka2o (jno, jo) = 0_il
                     nn_count = nn_count + 1
                  ENDIF
               ENDIF
            END DO
         END DO
         WRITE (nout,*) 'rem small a2o : ', nn_count
         !
      CASE ( 'o2a' )
         DO ja = 1, jpan
            zzs = SUM (wo2a (:, ja))
            DO jna = 1, jmo2a
               jo = ko2a (jna, ja)
               IF (jo > 0) THEN 
                  IF (wo2a (jna, ja) < epsfrac) THEN 
                     wo2a (jna, ja) = 0.0_rl
                     ko2a (jna, ja) = 0_il
                     nn_count = nn_count + 1
                  END IF
               END IF
            END DO
         END DO
         WRITE (nout,*) 'rem small o2a : ', nn_count
         !
      END SELECT
      !
      WRITE (nout,*) 'Appel de gather (', TRIM(clname),') dans rem_small'
      CALL gather (clname)
      !
   END SUBROUTINE rem_small
   !
END PROGRAM mosaic
!
